#include <stdlib.h>
#include <stdio.h>

#include "common.h"
#include <GL/freeglut.h>

#include "renderer.h"

/*****************************************************************************
 * mouse state                                                               *
 *****************************************************************************/

typedef struct 
{
  int  pos_x;
  int  pos_y;
  char btn_r;
  char btn_m;
  char btn_l;
} mouse_state_t;

static mouse_state_t mouse_state = {0};

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
static unsigned frame_counter = 0;

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
static void keyboard(unsigned char key,int x,int y)
{
  switch(key)
    {
    case 27:
      exit(0);
    case 'a':
      break;
    case 'b':
      break;
    case 'c':
      break;
    case 'd':
      break;
    default:
      break;
    }
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
static void mouse(int button, int state, int x, int y)
{
    if(button== GLUT_LEFT_BUTTON && state == GLUT_DOWN)
    {
	mouse_state.btn_l = 1;
	mouse_state.pos_x = x;
	mouse_state.pos_y = y;
    }
    if(button== GLUT_LEFT_BUTTON && state == GLUT_UP)
    {
    	mouse_state.btn_l = 0;
    }
    if(button== GLUT_RIGHT_BUTTON && state == GLUT_DOWN)
    {
	mouse_state.btn_r = 1;
	mouse_state.pos_x = x;
	mouse_state.pos_y = y;
    }
    if(button== GLUT_RIGHT_BUTTON && state == GLUT_UP)
    {
	mouse_state.btn_r = 0;
    }
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
static void motion(int x, int y)
{
  if(mouse_state.btn_l)
    {
      view_state.theta += y - mouse_state.pos_y;
      if(0 < view_state.theta)
        {
          view_state.theta = 0;
        }
      if(view_state.theta < -180)
        {
          view_state.theta = -180;
        }

      view_state.psi += x - mouse_state.pos_x;
      while(360 <= view_state.psi)
        {
	  view_state.psi -= 360;
        }
      while(view_state.psi < 0)
        {
	  view_state.psi += 360;
        }

      mouse_state.pos_x = x;
      mouse_state.pos_y = y;
    }

  if(mouse_state.btn_r)
    {
      view_state.distance += ((float)(y - mouse_state.pos_y)) / 10.0;
      if(view_state.distance < 1.0)
        {
          view_state.distance = 1.0;
        }
      if(40.0 < view_state.distance)
        {
          view_state.distance = 40.0;
        }
      mouse_state.pos_y = y;
    }

  if(mouse_state.btn_l || mouse_state.btn_r)
    {
      glutPostRedisplay();
    }
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
static void reshape(int w, int h)
{
  renderer_reshape(w, h);	
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
static void display(void)
{
  renderer_display();

  glutSwapBuffers();

  frame_counter++;
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
static void idle(void)
{
  glutPostRedisplay();
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
static void timer(int value)
{
  if(frame_counter) 
    {
      static char title[64];
      snprintf(title, sizeof(title), "%d Frames Per Second", frame_counter);
      glutSetWindowTitle(title);
    }
  frame_counter = 0;
  glutTimerFunc(1000, timer, value);
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
int main(int argc,char **argv)
{
  glutInit(&argc,argv);

  glutInitContextVersion(4, 0);
  glutInitContextFlags(GLUT_FORWARD_COMPATIBLE);
  glutInitContextProfile(GLUT_CORE_PROFILE);

  glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

  glutInitWindowSize(600,600);
  glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
  int window_handle = glutCreateWindow(argv[0]);
  if(window_handle < 1) 
    {
      fprintf(stderr, "ERROR: Could not create a new rendering window.\n");
      exit(EXIT_FAILURE);
    }

  glutDisplayFunc(display);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(keyboard);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutIdleFunc(idle);
  glutTimerFunc(0, timer, 0);

#if defined(USE_GLEW)
  GLenum glew_init_result = glewInit();
  if(GLEW_OK != glew_init_result) 
    {
      fprintf(stderr, "ERROR: %s\n", glewGetErrorString(glew_init_result));
      exit(EXIT_FAILURE);
    }
  fprintf(stdout, "OpenGL Version: %s\n", glGetString(GL_VERSION));
#endif

  int err;
  err = renderer_init();

  if(!err)
    {
      glutMainLoop();
    }

  renderer_kill();

  return err?EXIT_FAILURE:EXIT_SUCCESS;
}
