/* Local Variables: */
/* mode:c           */
/* End:             */
#version 400

in  vec4 ex_Position;
in  vec4 ex_Color;
in  vec3 ex_TexCoord;

out vec4 out_Color;

uniform sampler3D PerlinLookup;
uniform float     FrameCount;  

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
vec3 fade(vec3 t)
{ 
  return t * t * t * (t * (t * 6.0 - 15.0) + 10.0); 
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
float noise1(vec3 point)
{
  const vec3 tex_sz = textureSize(PerlinLookup, 0);
  const vec3 v000 = vec3(0, 0, 0);
  const vec3 v100 = vec3(1, 0, 0);
  const vec3 v010 = vec3(0, 1, 0);
  const vec3 v110 = vec3(1, 1, 0);
  const vec3 v001 = vec3(0, 0, 1);
  const vec3 v101 = vec3(1, 0, 1);
  const vec3 v011 = vec3(0, 1, 1);
  const vec3 v111 = vec3(1, 1, 1);

  /* FIND UNIT CUBE THAT CONTAINS POINT. */
  vec3 pi = floor(point);
  /* FIND RELATIVE X,Y,Z OF POINT IN CUBE. */
  vec3 pf = point - pi;
  /* COMPUTE FADE CURVES FOR EACH OF X,Y,Z. */
  vec3 fad = fade(pf); 
  /* HASH COORDINATES OF THE 8 CUBE CORNERS */
  vec3 g000 = texture(PerlinLookup, (pi + v000) / tex_sz).xyz * 4.0 - 1.0;
  vec3 g100 = texture(PerlinLookup, (pi + v100) / tex_sz).xyz * 4.0 - 1.0;
  vec3 g010 = texture(PerlinLookup, (pi + v010) / tex_sz).xyz * 4.0 - 1.0;
  vec3 g110 = texture(PerlinLookup, (pi + v110) / tex_sz).xyz * 4.0 - 1.0;
  vec3 g001 = texture(PerlinLookup, (pi + v001) / tex_sz).xyz * 4.0 - 1.0;
  vec3 g101 = texture(PerlinLookup, (pi + v101) / tex_sz).xyz * 4.0 - 1.0;
  vec3 g011 = texture(PerlinLookup, (pi + v011) / tex_sz).xyz * 4.0 - 1.0;
  vec3 g111 = texture(PerlinLookup, (pi + v111) / tex_sz).xyz * 4.0 - 1.0;

  float n000 = dot(g000, pf - v000);
  float n100 = dot(g100, pf - v100);
  float n010 = dot(g010, pf - v010);
  float n110 = dot(g110, pf - v110);
  float n001 = dot(g001, pf - v001);
  float n101 = dot(g101, pf - v101);
  float n011 = dot(g011, pf - v011);
  float n111 = dot(g111, pf - v111);

  /* AND ADD BLENDED RESULTS FROM 8 CORNERS OF CUBE */
  return mix(mix(mix(n000,
		     n100,
		     fad.x),
		 mix(n010,
		     n110,
		     fad.x),
		 fad.y),
	     mix(mix(n001,
		     n101,
		     fad.x),
		 mix(n011,
		     n111,
		     fad.x),
		 fad.y),
	     fad.z);
 
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
float perlin_turbulence(in vec3 vec, in float freq)
{
  float t;
  for (t = 0.0; freq >= 1.0; freq *= 0.5)
    {
      t += abs(noise1(freq * vec)) / freq;
    }
  return t;
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
void main(void)
{
  const float zoom = 2.0;
  const float freq = 3.9;

  vec3 point = 2.0 * ex_Position.xyz;
  float radius = length(point);

  float dr = perlin_turbulence(zoom * (point + vec3(0.0, 0.0, FrameCount)), freq);

  vec4 texel;
  texel.r = clamp(1.0 - 1.0 * pow(radius + dr / 3, 5), 0, 1);
  texel.g = clamp(1.0 - 2.0 * pow(radius + dr / 3, 3), 0, 1);
  texel.b = 0.0;
  texel.a = length(texel.rgb);
  
  out_Color = texel;
}
